"""Module for generating NPCs."""

import hashlib
import json
import os
import random
from collections import namedtuple

from strictyaml import Float, load, Map, Seq, Str


NPC = namedtuple(
    'NPC', [
        'name',
        'age',
        'race',
        'class_',
        'physical',
        'personality',
    ]
)

NPC_FILENAME = os.path.join(os.path.dirname(__file__), 'data/npc.yaml')
NPC_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data/npc.json')

NPC_SCHEMA = Map({
    'races': Seq(Map({'v': Str(), 'w': Float()})),
    'classes': Seq(Str()),
    'age': Seq(Map({'v': Str(), 'w': Float()})),
    'physical': Seq(Str()),
    'personality': Seq(Str()),
    'names': Seq(Str()),
})


def _read_data():
    """Read NPC data.

    Parsing StrictYAML turned out to be really slow -> The function attempts
    to read the pre-parsed JSON version. If that fails or if it has been
    generated for a different YAML data version, read the source YAML and try
    and store the data in JSON format for future runs.
    """

    def _handle_caching_error(operation, error):
        # Catch and ignore (just print warning) any error. This is just
        # "caching" and we want to continue with the rest in any case.
        print(
            f'WARNING: Could not {operation} the cache JSON file '
            f'("{NPC_JSON_FILENAME}"): {error}'
        )

    with open(NPC_FILENAME, 'r') as yaml_datafile:
        yaml_data = yaml_datafile.read()
        yaml_data_checksum = hashlib.sha512(yaml_data.encode()).hexdigest()

        # Try to read NPC_JSON_FILENAME ("cached"  version).
        try:
            with open(NPC_JSON_FILENAME, 'r') as json_datafile:
                npc_data = json.load(json_datafile)
                checksum = npc_data.pop('yaml_datafile_checksum')
                if checksum == yaml_data_checksum:
                    return npc_data
        except Exception as error:
            _handle_caching_error('read', error)

        # At this point we indeed need to parse the YAML data. Let's do so and
        # try and store the parsed data into a JSON file for future runs.
        npc_data = load(yaml_data, NPC_SCHEMA).data
        try:
            npc_data['yaml_datafile_checksum'] = yaml_data_checksum
            json.dump(npc_data, open(NPC_JSON_FILENAME, 'w'))
        except Exception as error:
            _handle_caching_error('store', error)
        return npc_data


NPC_DATA = _read_data()


def _weighted_random(data_set):
    """Returns a weighted random option from data_set.

    The data_set needs to be a sequence of dict-like objects with at
    least 'v' (value) and 'w' (weight) keys. Weights need to be
    convertible to float.
    """
    return random.choices(
        [x['v'] for x in data_set],
        [float(x['w']) for x in data_set]
    )[0]


def _filter_string_data(data_set, allowed=(), disallowed=()):
    """Filters the given weighted data according to given filters.

    The data_set needs to be a sequence of string values.
    The allowed and disallowed need to be sequences of strings and
    their contents will be used to filter the values in data_set.

    A data_set item will be a part of the result only if its value
    contains any of the allowed values as a substring unless it also
    contains any of the disallowed values as a substring.

    If allowed sequence is not provided, all items in data_set are
    allowed. If disallowed sequence is not provided, no items are
    filtered out.
    """
    if not allowed and not disallowed:
        return data_set

    if allowed:
        filtered = []
        for item in allowed:
            filtered += [x for x in data_set if item in x]
    else:
        filtered = data_set

    if disallowed:
        for item in disallowed:
            filtered = [x for x in filtered if item not in x]
    return filtered


def _filter_structured_data(data_set, allowed=(), disallowed=()):
    """Filters the given weighted data according to given filters.

    The data_set needs to be a sequence of dict-like objects with at
    least 'v' (value) key with string values.
    The allowed and disallowed need to be sequences of strings and
    their contents will be used to filter the values in data_set.

    A data_set item will be a part of the result only if its value
    contains any of the allowed values as a substring unless it also
    contains any of the disallowed values as a substring.

    If allowed sequence is not provided, all items in data_set are
    allowed. If disallowed sequence is not provided, no items are
    filtered out.
    """
    if not allowed and not disallowed:
        return data_set

    if allowed:
        filtered = []
        for item in allowed:
            filtered += [x for x in data_set if item in x['v']]
    else:
        filtered = data_set

    if disallowed:
        for item in disallowed:
            filtered = [x for x in filtered if item not in x['v']]
    return filtered


def generate_npc(detail_level=2, ages=None, classes=None, races=None):
    """Generate an NPC.

    Detail level affects how much detailed the generated NPC will be.
    Positive integer is expected, higher number means more details.
    Default is meant to give the best results, detail level of one gives
    just the bare minimum and with certain level of detail the NPCs tend
    to start getting contradictory traits (like fat and slim at the same
    time).

    Ages, classes and races are supposed to be sequences of allowed
    ages/classes/races. If None is passed instead of any of these,
    the default set of traits will be used.
    """
    age = str(_weighted_random(ages))

    if detail_level >= 2:
        class_ = str(random.choice(classes))  # nosec
    else:
        class_ = None

    name = str(random.choice(NPC_DATA['names']))  # nosec

    race = str(_weighted_random(races))

    physical = random.choices(NPC_DATA['physical'], k=detail_level)
    physical = [str(trait) for trait in physical]

    personality = random.choices(NPC_DATA['personality'], k=detail_level)
    personality = [str(trait) for trait in personality]

    return NPC(
        name=name,
        race=race,
        class_=class_,
        age=age,
        physical=physical,
        personality=personality,
    )


def generate_npcs(number=1, detail_level=2, filters=None):
    """Generate a number of NPCs.

    Detail level affects how much detailed the generated NPCs will be.
    Positive integer is expected, higher number means more details.
    Default is meant to give the best results, detail level of one gives
    just the bare minimum and with certain level of detail the NPCs tend
    to start getting contradictory traits (like fat and slim at the same
    time).

    Filters are expected to be a dictionary with string keys like
    'races_yes' and 'races_no' and sequence values with traits that are
    supposed to be included and excluded respectively while generating
    NPCs. Properties currently supporting filters are ages, classes and
    races.
    """
    # filter the data
    if filters is None:
        filters = {}
    ages = _filter_structured_data(NPC_DATA['age'],
                                   filters.get('ages_yes'),
                                   filters.get('ages_no'))
    # classes are not used for detail_level = 1
    if detail_level >= 2:
        classes = _filter_string_data(NPC_DATA['classes'],
                                      filters.get('classes_yes'),
                                      filters.get('classes_no'))
    else:
        classes = []
    races = _filter_structured_data(NPC_DATA['races'],
                                    filters.get('races_yes'),
                                    filters.get('races_no'))

    npcs = []
    for _ in range(number):
        npcs.append(generate_npc(detail_level, ages, classes, races))
    return npcs



#!/usr/bin/env python

"""Simple CLI tool for generating NPCs."""

import random
import sys

import click

from loreroll.npc import generate_npcs


def print_npc(npc):
    """Print the given NPC."""
    print(f'Name: {npc.name}')
    print(f'Age: {npc.age}')
    print(f'Race: {npc.race}')
    if npc.class_:
        print(f'Class: {npc.class_}')
    print(f'Appearance: {", ".join(npc.physical)}')
    print(f'Personality: {", ".join(npc.personality)}')
    print()


# pylint: disable=too-many-arguments
@click.command()
@click.option('--age-allowed', '-a', 'ages_yes', multiple=True,
              help='Allowed age(s).')
@click.option('--age-disallowed', '-A', 'ages_no', multiple=True,
              help='Disallowed age(s).')
@click.option('--class-allowed', '-c', 'classes_yes', multiple=True,
              help='Allowed class(es).')
@click.option('--class-disallowed', '-C', 'classes_no', multiple=True,
              help='Disallowed class(es).')
@click.option('--detail-level', '-d', 'detail_level', default=2,
              help='Amount of details generated (one or higher).')
@click.option('--number', '-n', default=1,
              help='Number of NPCs to generate.')
@click.option('--race-allowed', '-r', 'races_yes', multiple=True,
              help='Allowed race(s).')
@click.option('--race-disallowed', '-R', 'races_no', multiple=True,
              help='Disallowed race(s).')
@click.option('--seed', '-s', 'seed', default=None,
              help='Seed number used to generate NPCs. The same seed will '
                   'produce the same results.')
def generate(number=1, ages_yes=None, ages_no=None, classes_yes=None,
             classes_no=None, detail_level=2, races_yes=None,
             races_no=None, seed=None):
    """Generate 'number' of NPCs and print them."""
    filters = {
        'ages_no': ages_no,
        'ages_yes': ages_yes,
        'classes_no': classes_no,
        'classes_yes': classes_yes,
        'races_no': races_no,
        'races_yes': races_yes,
    }

    # Seed properly - use either the value from command line or seed randomly.
    if seed is None:
        # String needed instead of int because command line options are also
        # strings.
        seed = str(random.randrange(sys.maxsize))  # nosec
    random.seed(seed)
    print(f"Seed used: '{seed}'. Run with '-s {seed}' to get the same "
          f"result.\n")

    npcs = generate_npcs(number, detail_level=detail_level, filters=filters)
    for npc in npcs:
        print_npc(npc)
# pylint: enable=too-many-arguments


if __name__ == '__main__':
    generate()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['loreroll']

package_data = \
{'': ['*'], 'loreroll': ['data/*']}

modules = \
['rollnpc']
install_requires = \
['click>=7.1.2,<7.2.0', 'strictyaml>=1.0.6,<1.2.0']

entry_points = \
{'console_scripts': ['rollnpc = rollnpc:generate']}

setup_kwargs = {
    'name': 'rollthelore',
    'version': '0.2',
    'description': 'An unseen servant providing an advantage to DMs/GMs while creating their worlds.',
    'long_description': "# RollTheLore\nAn *unseen servant* providing an *advantage* to DMs/GMs while creating their worlds.\n\n[![Build Status](https://travis-ci.com/geckon/rollthelore.svg?branch=master)](https://travis-ci.com/geckon/rollthelore)\n\nAs a DM, have you ever needed to lower the *DC* for the world building skill\ncheck? Did you need to create the right NPC for your players to interact with\nand thought you could use a *divine intervention*? Were you out of ideas and\nfelt like an *inspiration die* or a bit of *luck* was all that you needed? This\ntool will probably not fulfill all your *wish*es but it can at least provide\n*guidance*.\n\nRollTheLore is a tool from a DM to other DMs out there but it can also help\nplayers as it's supposed to inspire you while not only creating your world,\na shop or a random NPC encounter but it can also be used to help with your\nbackstories.\n\nAs of now it can only randomly create NPCs but I have some ideas for generating\ntowns as well and who knows? Maybe there will be even more. RollTheLore is not\nmeant to be a perfect tool but rather a simple one that still can provide very\nvaluable ideas for your campaigns and stories. At least for me personaly it\nworks very well as I often need a few simple points I can use as inspiration\nand build more fluff around it with both imagination and improvisation.\n\nSometimes the generated character doesn't make much sense but often that is\nexactly the fun part. When I need an NPC for a story I like to *roll* a few of\nthem and pick one as a basis, then I build on that. It can also be used to\npre-roll a few NPCs in advance and then use them e.g. when your players decide\nto enter a shop or address a person you hadn't planned beforehand.\n\nPrimarily RollTheLore is intended to be used with DnD 5e but it can very well\nserve for other game systems as well. The imagination is what matters the most.\n\nPlease note that the tool is under development. Ideas, comments and bug reports are\nwelcome!\n\n## Installation\n\n```\npip install rollthelore\n```\n\n## Usage\n\n```\n$ rollnpc --help\nUsage: rollnpc [OPTIONS]\n\n  Generate 'number' of NPCs and print them.\n\nOptions:\n  -a, --age-allowed TEXT       Allowed age(s).\n  -A, --age-disallowed TEXT    Disallowed age(s).\n  -c, --class-allowed TEXT     Allowed class(es).\n  -C, --class-disallowed TEXT  Disallowed class(es).\n  -d, --detail-level INTEGER   Amount of details generated (one or higher).\n  -n, --number INTEGER         Number of NPCs to generate.\n  -r, --race-allowed TEXT      Allowed race(s).\n  -R, --race-disallowed TEXT   Disallowed race(s).\n  --help                       Show this message and exit.\n```\n\n## Examples\n\n```\n$ rollnpc\nName: Zaniel\nAge: older\nRace: tabaxi\nClass: monk\nAppearance: deaf, blind\nPersonality: discouraging, disrespectful\n```\n\n```\n$ rollnpc -n3\nName: Iseult\nAge: old\nRace: dragonborn\nClass: fighter\nAppearance: cain, moustache\nPersonality: courteous, foolish\n\nName: Nyssa\nAge: adult\nRace: dragonborn\nClass: barbarian\nAppearance: hook instead of a hand, underbite\nPersonality: miserable, suspicious\n\nName: Briallan\nAge: older\nRace: human\nClass: wizard\nAppearance: skinny, tall\nPersonality: honorable, stylish\n```\n\n```\n$ rollnpc -n2 -r elf\nName: Evadne\nAge: young\nRace: elf (wood)\nClass: ranger\nAppearance: silent voice, cute\nPersonality: dependable, honorable\n\nName: Ianthe\nAge: young\nRace: half-elf\nClass: warlock\nAppearance: handlebar beard, fire-burnt skin\nPersonality: witty, organized\n```\n\n```\n$ rollnpc --detail-level 1\nName: Séverin\nAge: young\nRace: tiefling\nAppearance: mute\nPersonality: happy\n```\n\n```\n$ rollnpc --detail-level 3\nName: Korbin\nAge: adult\nRace: gnome (forest)\nClass: ranger\nAppearance: big nose, scarred, sinewy\nPersonality: grim, regretful, naive\n\n```\n\n### Seeding\n\nLet's say you generated this lovely duo and you want to keep it for the future.\n\n```\n$ rollnpc.py -n2\nSeed used: '6095344300345411392'. Run with '-s 6095344300345411392' to get the same result.\n\nName: Macon\nAge: older\nRace: half-elf\nClass: bard\nAppearance: big eyes, muttonchops\nPersonality: intellectual, secretive\n\nName: Sirius\nAge: very old\nRace: human\nClass: cleric\nAppearance: different hand size, dimple in chin\nPersonality: speaks silently, hypochondriac\n```\n\nYou can either save the whole text or just the seed and generate the same\ndata again like this:\n\n```\n$ rollnpc.py -n2 -s 6095344300345411392\nSeed used: '6095344300345411392'. Run with '-s 6095344300345411392' to get the same result.\n\nName: Macon\nAge: older\nRace: half-elf\nClass: bard\nAppearance: big eyes, muttonchops\nPersonality: intellectual, secretive\n\nName: Sirius\nAge: very old\nRace: human\nClass: cleric\nAppearance: different hand size, dimple in chin\nPersonality: speaks silently, hypochondriac\n```\n",
    'author': 'Tomáš Heger',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
